# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
import os
import socket
import subprocess

import numpy as np
import torch
import torch.distributed as dist

from ..commands.run import _run
from ..outputs import create_output
from ..runners import create_runner
from . import runner_registry
from .default import DefaultRunner

LOG = logging.getLogger(__name__)


def create_parallel_runner(config, pid):
    runner = create_runner(config, pid=pid)
    _run(runner, config)


@runner_registry.register("parallel")
class ParallelRunner(DefaultRunner):
    """Runner which splits a model over multiple devices"""

    def __init__(self, context, pid=0):
        super().__init__(context)

        self.model_comm_group = None
        self.pid = pid

        self._bootstrap_processes()

        # disable most logging on non-zero ranks
        if self.global_rank != 0:
            logging.getLogger().setLevel(logging.WARNING)

        if self.device == "cuda":
            self.device = f"{self.device}:{self.local_rank}"
            torch.cuda.set_device(self.local_rank)

        # Create a model comm group for parallel inference
        # A dummy comm group is created if only a single device is in use
        if self.world_size > 1:
            model_comm_group = self._init_parallel()
            self.model_comm_group = model_comm_group

            # Ensure each parallel model instance uses the same seed
            self._seed_procs()
        else:
            LOG.warning("ParallelRunner selected but world size of 1 detected")

    def predict_step(self, model, input_tensor_torch, fcstep, **kwargs):
        if self.model_comm_group is None:
            return model.predict_step(input_tensor_torch)
        else:
            try:
                return model.predict_step(input_tensor_torch, self.model_comm_group)
            except TypeError as err:
                LOG.error(
                    "Please upgrade to a newer version of anemoi-models (at least version v0.4.2) to use parallel inference. If updating breaks your checkpoints, you can try reverting to your original version of anemoi-models and cherry-picking 'https://github.com/ecmwf/anemoi-core/pull/77'"
                )
                raise err

    def create_output(self):
        if self.global_rank == 0:
            output = create_output(self, self.config.output)
            LOG.info("Output: %s", output)
            return output
        else:
            output = create_output(self, "none")
            return output

    def __del__(self):
        if self.model_comm_group is not None:
            dist.destroy_process_group()

    def _seed_procs(self):
        """Ensures each process uses the same seed.
        Will try read 'ANEMOI_BASE_SEED' from the environment
        Otherwise, the seed of process 0 will be shared to all processes
        """

        seed = None
        seed_threshold = 1000
        env_var_list = ["ANEMOI_BASE_SEED"]
        for env_var in env_var_list:
            if env_var in os.environ:
                seed = int(os.environ.get(env_var))
                if seed < seed_threshold:
                    seed *= seed_threshold  # make it (hopefully) big enough
                break

        if self.global_rank == 0:
            if seed is None:
                seed = torch.initial_seed()
            torch.distributed.broadcast_object_list([seed], src=0, group=self.model_comm_group)
        else:
            msg_buffer = np.array([1], dtype=np.uint64)
            torch.distributed.broadcast_object_list(msg_buffer, src=0, group=self.model_comm_group)
            seed = msg_buffer[0]
            torch.manual_seed(seed)

    def _srun_used(self):
        """returns true if anemoi-inference was launched with srun"""

        # from pytorch lightning
        # https://github.com/Lightning-AI/pytorch-lightning/blob/a944e7744e57a5a2c13f3c73b9735edf2f71e329/src/lightning/fabric/plugins/environments/slurm.py
        return "SLURM_NTASKS" in os.environ and os.environ.get("SLURM_JOB_NAME") not in ("bash", "interactive")

    def _spawn_parallel_procs(self, num_procs):
        """When srun is not available, this method creates N-1 child processes within the same node for parallel inference"""
        LOG.debug(f"spawning {num_procs -1 } procs")

        # check num_procs <= num_gpus
        if self.device.startswith("cuda"):
            num_gpus = torch.cuda.device_count()
            if num_procs > num_gpus:
                raise ValueError(
                    f"You requested parallel inference over {num_procs} GPUs but your node only has {num_gpus} GPUs available."
                )

        # Create N-1 procs, each with a unique PID
        import torch.multiprocessing as mp

        mp.set_start_method("spawn")
        config = self.config
        for pid in range(1, num_procs):
            mp.Process(target=create_parallel_runner, args=(config, pid)).start()

    def _bootstrap_processes(self):
        """initalises processes and their network information
        If srun is available, slurm variables are read to determine network settings.
        Otherwise, local processes are spawned and network info is infered from config
        """
        using_slurm = self._srun_used()
        if using_slurm:

            # Determine world size and rank from slurm env vars
            global_rank, local_rank, world_size = self._get_parallel_info_from_slurm()
            self.global_rank = global_rank
            self.local_rank = local_rank
            self.world_size = world_size

            # determine master address and port from slurm/override env vars
            slurm_addr, slurm_port = self._init_network_from_slurm()
            self.master_addr = slurm_addr
            self.master_port = slurm_port

            # the world size entry in the config is only needed when not launching via srun
            if self.config.world_size != 1:
                LOG.warning(
                    f"world size ({self.config.world_size}) set in the config is ignored because we are launching via srun, using 'SLURM_NTASKS' instead"
                )
        else:
            # If srun is not available, spawn procs manually on a node

            # Read the config to determine world_size and pid
            self.global_rank = self.pid  # only inference within a single node is supported when not using srun
            self.local_rank = self.pid
            self.world_size = self.config.world_size
            if self.world_size == 1:
                LOG.warning(
                    "You selected 'runner: parallel' but you have only set 'world_size: 1'. Please update world_size or launch via srun to make use of parallel inference"
                )
            if self.world_size <= 0:
                raise ValueError(
                    f"Error. 'world_size' must be greater then 1 to use parallel inference. {world_size=} set in the config is invalid."
                )

            # since we are running within a node, 'localhost' and any port can be used
            self.master_addr = "localhost"
            # generates a port between 10000 and 19999, based on the nodes hostname (which will be the same across all node-local procs)
            import hashlib

            node_name = os.uname().nodename.encode()  # Convert to bytes
            hash_val = int(hashlib.md5(node_name).hexdigest(), 16)  # Convert hash to int
            self.master_port = 10000 + (hash_val % 9999)

            # Spawn the other processes manually
            if self.local_rank == 0:
                self._spawn_parallel_procs(self.world_size)

    def _init_network_from_slurm(self):
        """Reads Slurm environment to set master address and port for parallel communication"""

        # Get the master address from the SLURM_NODELIST environment variable
        slurm_nodelist = os.environ.get("SLURM_NODELIST")
        if not slurm_nodelist:
            raise ValueError("SLURM_NODELIST environment variable is not set.")

        # Check if MASTER_ADDR is given, otherwise try set it using 'scontrol'
        master_addr = os.environ.get("MASTER_ADDR")
        if master_addr is None:
            LOG.debug("'MASTER_ADDR' environment variable not set. Trying to set via SLURM")
            try:
                result = subprocess.run(
                    ["scontrol", "show", "hostname", slurm_nodelist], stdout=subprocess.PIPE, text=True, check=True
                )
            except subprocess.CalledProcessError as err:
                LOG.error(
                    "Python could not execute 'scontrol show hostname $SLURM_NODELIST' while calculating MASTER_ADDR. You could avoid this error by setting the MASTER_ADDR env var manually."
                )
                raise err

            master_addr = result.stdout.splitlines()[0]

            # Resolve the master address using nslookup
            try:
                master_addr = socket.gethostbyname(master_addr)
            except socket.gaierror:
                raise ValueError(f"Could not resolve hostname: {master_addr}")

        # Check if MASTER_PORT is given, otherwise generate one based on SLURM_JOBID
        master_port = os.environ.get("MASTER_PORT")
        if master_port is None:
            LOG.debug("'MASTER_PORT' environment variable not set. Trying to set via SLURM")
            slurm_jobid = os.environ.get("SLURM_JOBID")
            if not slurm_jobid:
                raise ValueError("SLURM_JOBID environment variable is not set.")

            master_port = str(10000 + int(slurm_jobid[-4:]))

        # Print the results for confirmation
        LOG.debug(f"MASTER_ADDR: {master_addr}")
        LOG.debug(f"MASTER_PORT: {master_port}")

        return master_addr, master_port

    def _init_parallel(self):
        """Creates a model communication group to be used for parallel inference"""

        if self.world_size > 1:

            # use 'startswith' instead of '==' in case device is 'cuda:0'
            if self.device.startswith("cuda"):
                backend = "nccl"
            else:
                if dist.is_mpi_available():
                    backend = "mpi"
                else:
                    backend = "gloo"

            dist.init_process_group(
                backend=backend,
                init_method=f"tcp://{self.master_addr}:{self.master_port}",
                timeout=datetime.timedelta(minutes=3),
                world_size=self.world_size,
                rank=self.global_rank,
            )
            LOG.info(f"Creating a model communication group with {self.world_size} devices with the {backend} backend")

            model_comm_group_ranks = np.arange(self.world_size, dtype=int)
            model_comm_group = dist.new_group(model_comm_group_ranks)
        else:
            model_comm_group = None

        return model_comm_group

    def _get_parallel_info_from_slurm(self):
        """Reads Slurm env vars, if they exist, to determine if inference is running in parallel"""
        local_rank = int(os.environ.get("SLURM_LOCALID", 0))  # Rank within a node, between 0 and num_gpus
        global_rank = int(os.environ.get("SLURM_PROCID", 0))  # Rank within all nodes
        world_size = int(os.environ.get("SLURM_NTASKS", 1))  # Total number of processes

        return global_rank, local_rank, world_size



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
from abc import ABC
from abc import abstractmethod

LOG = logging.getLogger(__name__)

# TODO: only one method is need: `load_data`.
# The other methods can be implemenneted concreetly
# using the `load_data` method.


class Input(ABC):
    """_summary_"""

    def __init__(self, context):
        self.context = context
        self.checkpoint = context.checkpoint

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @abstractmethod
    def create_input_state(self, *, date=None):
        """Create the input state dictionary."""
        pass

    @abstractmethod
    def load_forcings(self, *, variables, dates):
        """Load forcings (constant and dynamic)."""
        pass

    def input_variables(self):
        """Return the list of input variables"""
        return list(self.checkpoint.variable_to_input_tensor_index.keys())

    def set_private_attributes(self, state, value):
        """Provide a way to a subclass to set private attributes in the state
        dictionary, that may be needed but the output object.
        """
        pass

    def template(self, variable, date, **kwargs):
        """Used for fetching GRIB templates."""
        raise NotImplementedError(f"{self.__class__.__name__}.template() not implemented")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from ..forcings import ComputedForcings
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)


@runner_registry.register("plugin")
class PluginRunner(Runner):
    """A runner implementing the ai-models plugin API."""

    def __init__(self, checkpoint: str, *, device: str):
        super().__init__(checkpoint, device=device)

    # Compatibility with the ai_models API

    @property
    def param_sfc(self):
        params, _ = self.checkpoint.mars_by_levtype("sfc")
        return sorted(params)

    @property
    def param_level_pl(self):
        params, levels = self.checkpoint.mars_by_levtype("pl")
        return sorted(params), sorted(levels)

    @property
    def param_level_ml(self):
        params, levels = self.checkpoint.mars_by_levtype("ml")
        return sorted(params), sorted(levels)

    @property
    def lagged(self):
        return [s.total_seconds() // 3600 for s in self.checkpoint.lagged]

    def create_constant_computed_forcings(self, variables, mask):
        result = ComputedForcings(self, variables, mask)
        LOG.info("Constant computed forcing: %s", result)
        return result

    def create_dynamic_computed_forcings(self, variables, mask):
        result = ComputedForcings(self, variables, mask)
        LOG.info("Dynamic computed forcing: %s", result)
        return result



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging

import numpy as np

from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


def print_state(state, print=print):
    print()
    print("😀", end=" ")
    for key, value in state.items():

        if isinstance(value, datetime.datetime):
            print(f"{key}={value.isoformat()}", end=" ")

        if isinstance(value, (str, float, int, bool, type(None))):
            print(f"{key}={value}", end=" ")

        if isinstance(value, np.ndarray):
            print(f"{key}={value.shape}", end=" ")

    fields = state.get("fields", {})

    print(f"fields={len(fields)}")
    print()

    names = list(fields.keys())
    n = 4

    idx = list(range(0, len(names), max(1, len(names) // n)))
    idx.append(len(names) - 1)
    idx = sorted(set(idx))

    length = max(len(name) for name in names)

    for i in idx:
        name = names[i]
        field = fields[name]
        min_value = f"min={np.nanmin(field):g}"
        max_value = f"max={np.nanmax(field):g}"
        print(f"    {name:{length}} shape={field.shape} {min_value:18s} {max_value:18s}")

    print()


@output_registry.register("printer")
class PrinterOutput(Output):
    """_summary_"""

    def write_step(self, state):
        print_state(state)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
import warnings
from functools import cached_property

import numpy as np
import torch
from anemoi.utils.dates import frequency_to_timedelta as to_timedelta
from anemoi.utils.text import table
from anemoi.utils.timer import Timer  # , Timers

from .checkpoint import Checkpoint
from .context import Context
from .postprocess import Accumulator
from .postprocess import Noop
from .precisions import PRECISIONS

LOG = logging.getLogger(__name__)


class Kind:
    """Used for debugging purposes"""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __repr__(self):
        result = []

        for k, v in self.kwargs.items():
            if v:
                result.append(k)

        if not result:
            return "?"

        return ", ".join(result)


class Runner(Context):
    """A runner is responsible for running a model."""

    def __init__(
        self,
        checkpoint,
        *,
        accumulate_from_start_of_forecast=False,
        device: str = "cuda",
        precision: str = None,
        report_error=False,
        allow_nans=None,  # can be True of False
        use_grib_paramid=False,
        verbosity=0,
        inference_options=None,
        patch_metadata={},
        development_hacks={},  # For testing purposes, don't use in production
        output_frequency=None,
        write_initial_state=True,
    ):
        self._checkpoint = Checkpoint(checkpoint, patch_metadata=patch_metadata)

        self.device = device
        self.precision = precision
        self.report_error = report_error

        # Override the default values set in `Context`
        self.verbosity = verbosity
        self.allow_nans = allow_nans
        self.use_grib_paramid = use_grib_paramid
        self.development_hacks = development_hacks
        self.hacks = bool(development_hacks)
        self.output_frequency = output_frequency
        self.write_initial_state = write_initial_state

        # This could also be passed as an argument

        self.postprocess = Noop()

        if accumulate_from_start_of_forecast is True:
            # Get accumulations from the checkpoint
            accumulations = self.checkpoint.accumulations

            if accumulations:
                self.postprocess = Accumulator(accumulations)

        self._input_kinds = {}
        self._input_tensor_by_name = []

        self._output_kinds = {}
        self._output_tensor_by_name = []

        if self.verbosity > 2:
            self.checkpoint.print_indices()

        LOG.info("Using %s runner", self.__class__.__name__)

        if self.verbosity > 1:
            self.checkpoint.print_variable_categories()

    @property
    def checkpoint(self):
        return self._checkpoint

    def run(self, *, input_state, lead_time):

        self.constant_forcings_inputs = self.checkpoint.constant_forcings_inputs(self, input_state)
        self.dynamic_forcings_inputs = self.checkpoint.dynamic_forcings_inputs(self, input_state)
        self.boundary_forcings_inputs = self.checkpoint.boundary_forcings_inputs(self, input_state)

        # timers = Timers()

        lead_time = to_timedelta(lead_time)

        # This may be used but Output objects to compute the step
        self.lead_time = lead_time
        self.time_step = self.checkpoint.timestep

        input_tensor = self.prepare_input_tensor(input_state)

        try:
            yield from self.postprocess(self.forecast(lead_time, input_tensor, input_state))
        except (TypeError, ModuleNotFoundError, AttributeError):
            if self.report_error:
                self.checkpoint.report_error()
            raise

        # timers.report()

    def add_initial_forcings_to_input_state(self, input_state):
        # Should that be alreay a list of dates
        date = input_state["date"]
        fields = input_state["fields"]

        dates = [date + h for h in self.checkpoint.lagged]

        # For output object. Should be moved elsewhere
        self.reference_date = dates[-1]
        self.initial_dates = dates

        # TODO: Check for user provided forcings

        for source in self.constant_forcings_inputs:
            LOG.info("Constant forcings input: %s %s (%s)", source, source.variables, dates)
            arrays = source.load_forcings(input_state, dates)
            for name, forcing in zip(source.variables, arrays):
                assert isinstance(forcing, np.ndarray), (name, forcing)
                fields[name] = forcing
                self._input_kinds[name] = Kind(forcing=True, constant=True, **source.kinds)

        for source in self.dynamic_forcings_inputs:
            LOG.info("Dynamic forcings input: %s %s (%s)", source, source.variables, dates)
            arrays = source.load_forcings(input_state, dates)
            for name, forcing in zip(source.variables, arrays):
                assert isinstance(forcing, np.ndarray), (name, forcing)
                fields[name] = forcing
                self._input_kinds[name] = Kind(forcing=True, constant=True, **source.kinds)

    def prepare_input_tensor(self, input_state, dtype=np.float32):

        if "latitudes" not in input_state:
            input_state["latitudes"] = self.checkpoint.latitudes

        if "longitudes" not in input_state:
            input_state["longitudes"] = self.checkpoint.longitudes

        if input_state.get("latitudes") is None or input_state.get("longitudes") is None:
            raise ValueError("Input state must contain 'latitudes' and 'longitudes'")

        typed_variables = self.checkpoint.typed_variables

        for name in input_state["fields"]:
            self._input_kinds[name] = Kind(input=True, constant=typed_variables[name].is_constant_in_time)

        # Add initial forcings to input state if needed

        self.add_initial_forcings_to_input_state(input_state)

        input_state = self.validate_input_state(input_state)

        input_fields = input_state["fields"]

        input_tensor_numpy = np.full(
            shape=(
                self.checkpoint.multi_step_input,
                self.checkpoint.number_of_input_features,
                self.checkpoint.number_of_grid_points,
            ),
            fill_value=np.nan,
            dtype=dtype,
        )

        self._input_tensor_by_name = [None] * self.checkpoint.number_of_input_features

        LOG.info("Preparing input tensor with shape %s", input_tensor_numpy.shape)

        variable_to_input_tensor_index = self.checkpoint.variable_to_input_tensor_index

        check = set()
        for var, field in input_fields.items():
            i = variable_to_input_tensor_index[var]
            if i in check:
                raise ValueError(f"Duplicate variable {var}/{i} in input fields")
            input_tensor_numpy[:, i] = field
            check.add(i)

            self._input_tensor_by_name[i] = var

        if len(check) != self.checkpoint.number_of_input_features:
            missing = set(range(self.checkpoint.number_of_input_features)) - check
            mapping = {v: k for k, v in self.checkpoint.variable_to_input_tensor_index.items()}
            raise ValueError(f"Missing variables in input fields: {[mapping.get(_,_) for _ in missing]}")

        return input_tensor_numpy

    @cached_property
    def autocast(self):
        autocast = self.precision

        if autocast is None:
            autocast = self.checkpoint.precision

        if autocast is None:
            LOG.warning("No autocast given, using float16")
            autocast = "16"

        return PRECISIONS.get(autocast, autocast)

    @cached_property
    def model(self):
        with Timer(f"Loading {self.checkpoint}"):
            model = torch.load(self.checkpoint.path, map_location=self.device, weights_only=False).to(self.device)
            # model.set_inference_options(**self.inference_options)
            return model

    def predict_step(self, model, input_tensor_torch, fcstep, **kwargs):
        # extra args are only used in specific runners
        # TODO: move this to a Stepper class.
        return model.predict_step(input_tensor_torch)

    def forecast(self, lead_time, input_tensor_numpy, input_state):
        self.model.eval()

        torch.set_grad_enabled(False)

        # Create pytorch input tensor
        input_tensor_torch = torch.from_numpy(np.swapaxes(input_tensor_numpy, -2, -1)[np.newaxis, ...]).to(self.device)

        lead_time = to_timedelta(lead_time)
        steps = lead_time // self.checkpoint.timestep

        LOG.info("Using autocast %s", self.autocast)
        LOG.info("Lead time: %s, time stepping: %s Forecasting %s steps", lead_time, self.checkpoint.timestep, steps)

        result = input_state.copy()  # We should not modify the input state
        result["fields"] = dict()

        start = input_state["date"]

        # The variable `check` is used to keep track of which variables have been updated
        # In the input tensor. `reset` is used to reset `check` to False except
        # when the values are of the constant in time variables

        reset = np.full((input_tensor_torch.shape[-1],), False)
        variable_to_input_tensor_index = self.checkpoint.variable_to_input_tensor_index
        typed_variables = self.checkpoint.typed_variables
        for variable, i in variable_to_input_tensor_index.items():
            if typed_variables[variable].is_constant_in_time:
                reset[i] = True

        check = reset.copy()

        if self.verbosity > 0:
            self._print_input_tensor("First input tensor", input_tensor_torch)

        for s in range(steps):
            step = (s + 1) * self.checkpoint.timestep
            date = start + step
            LOG.info("Forecasting step %s (%s)", step, date)

            result["date"] = date

            # Predict next state of atmosphere
            with torch.autocast(device_type=self.device, dtype=self.autocast):
                y_pred = self.predict_step(self.model, input_tensor_torch, fcstep=s)

            # Detach tensor and squeeze (should we detach here?)
            output = np.squeeze(y_pred.cpu().numpy())  # shape: (values, variables)

            # Update state
            for i in range(output.shape[1]):
                result["fields"][self.checkpoint.output_tensor_index_to_variable[i]] = output[:, i]

            if (s == 0 and self.verbosity > 0) or self.verbosity > 1:
                self._print_output_tensor("Output tensor", output)

            yield result

            # No need to prepare next input tensor if we are at the last step
            if s == steps - 1:
                continue

            # Update  tensor for next iteration

            check[:] = reset

            input_tensor_torch = self.copy_prognostic_fields_to_input_tensor(input_tensor_torch, y_pred, check)

            del y_pred  # Recover memory

            input_tensor_torch = self.add_dynamic_forcings_to_input_tensor(input_tensor_torch, input_state, date, check)
            input_tensor_torch = self.add_boundary_forcings_to_input_tensor(
                input_tensor_torch, input_state, date, check
            )

            if not check.all():
                # Not all variables have been updated
                missing = []
                variable_to_input_tensor_index = self.checkpoint.variable_to_input_tensor_index
                mapping = {v: k for k, v in variable_to_input_tensor_index.items()}
                for i in range(check.shape[-1]):
                    if not check[i]:
                        missing.append(mapping[i])

                raise ValueError(f"Missing variables in input tensor: {sorted(missing)}")

            if (s == 0 and self.verbosity > 0) or self.verbosity > 1:
                self._print_input_tensor("Next input tensor", input_tensor_torch)

    def copy_prognostic_fields_to_input_tensor(self, input_tensor_torch, y_pred, check):

        # input_tensor_torch is shape: (batch, multi_step_input, values, variables)
        # batch is always 1

        prognostic_output_mask = self.checkpoint.prognostic_output_mask
        prognostic_input_mask = self.checkpoint.prognostic_input_mask

        # Copy prognostic fields to input tensor
        prognostic_fields = y_pred[..., prognostic_output_mask]  # Get new predicted values
        input_tensor_torch = input_tensor_torch.roll(-1, dims=1)  # Roll the tensor in the multi_step_input dimension
        input_tensor_torch[:, -1, :, self.checkpoint.prognostic_input_mask] = (
            prognostic_fields  # Add new values to last 'multi_step_input' row
        )

        assert not check[prognostic_input_mask].any()  # Make sure we are not overwriting some values
        check[prognostic_input_mask] = True

        for n in prognostic_input_mask:
            self._input_kinds[self._input_tensor_by_name[n]] = Kind(prognostic=True)

        return input_tensor_torch

    def add_dynamic_forcings_to_input_tensor(self, input_tensor_torch, state, date, check):

        if self.hacks:
            if "dynamic_forcings_date" in self.development_hacks:
                date = self.development_hacks["dynamic_forcings_date"]
                warnings.warn(f"🧑‍💻 Using `dynamic_forcings_date` hack: {date} 🧑‍💻")

        # TODO: check if there were not already loaded as part of the input state

        # input_tensor_torch is shape: (batch, multi_step_input, values, variables)
        # batch is always 1

        for source in self.dynamic_forcings_inputs:

            forcings = source.load_forcings(state, [date])  # shape: (variables, dates, values)

            forcings = np.squeeze(forcings, axis=1)  # Drop the dates dimension

            forcings = np.swapaxes(forcings[np.newaxis, np.newaxis, ...], -2, -1)  # shape: (1, 1, values, variables)

            forcings = torch.from_numpy(forcings).to(self.device)  # Copy to device

            input_tensor_torch[:, -1, :, source.mask] = forcings  # Copy forcings to last 'multi_step_input' row

            assert not check[source.mask].any()  # Make sure we are not overwriting some values
            check[source.mask] = True

            for n in source.mask:
                self._input_kinds[self._input_tensor_by_name[n]] = Kind(forcing=True, **source.kinds)

        return input_tensor_torch

    def add_boundary_forcings_to_input_tensor(self, input_tensor_torch, state, date, check):

        # input_tensor_torch is shape: (batch, multi_step_input, values, variables)
        # batch is always 1
        sources = self.boundary_forcings_inputs
        for source in sources:
            forcings = source.load_forcings(state, [date])  # shape: (variables, dates, values)

            forcings = np.squeeze(forcings, axis=1)  # Drop the dates dimension

            forcings = np.swapaxes(forcings[np.newaxis, np.newaxis, ...], -2, -1)  # shape: (1, 1, values, variables)
            forcings = torch.from_numpy(forcings).to(self.device)  # Copy to device
            total_mask = np.ix_([0], [-1], source.spatial_mask, source.variables_mask)
            input_tensor_torch[total_mask] = forcings  # Copy forcings to last 'multi_step_input' row

        # TO DO: add some consistency checks as above
        return input_tensor_torch

    def validate_input_state(self, input_state):

        if not isinstance(input_state, dict):
            raise ValueError("Input state must be a dictionnary")

        EXPECT = dict(date=datetime.datetime, latitudes=np.ndarray, longitudes=np.ndarray, fields=dict)

        for key, klass in EXPECT.items():
            if key not in input_state:
                raise ValueError(f"Input state must contain a `{key}` enytry")

            if not isinstance(input_state[key], klass):
                raise ValueError(
                    f"Input state entry `{key}` is type {type(input_state[key])}, expected {klass} instead"
                )

        # Detach from the user's input so we can modify it
        input_state = input_state.copy()
        fields = input_state["fields"] = input_state["fields"].copy()
        number_of_grid_points = self.checkpoint.number_of_grid_points

        for latlon in ("latitudes", "longitudes"):
            if len(input_state[latlon].shape) != 1:
                raise ValueError(f"Input state entry `{latlon}` must be 1D, shape is {input_state[latlon].shape}")

        nlat = len(input_state["latitudes"])
        nlon = len(input_state["longitudes"])
        if nlat != nlon:
            raise ValueError(f"Size mismatch latitudes={nlat}, longitudes={nlon}")

        if nlat != number_of_grid_points:
            raise ValueError(f"Size mismatch latitudes={nlat}, number_of_grid_points={number_of_grid_points}")

        multi_step = self.checkpoint.multi_step_input

        expected_shape = (multi_step, number_of_grid_points)

        LOG.info("Expected shape for each input fields: %s", expected_shape)

        # Check field
        with_nans = []

        for name, field in list(fields.items()):

            # Allow for 1D fields if multi_step is 1
            if len(field.shape) == 1:
                field = fields[name] = field.reshape(1, field.shape[0])

            if field.shape != expected_shape:
                raise ValueError(f"Field `{name}` has the wrong shape. Expected {expected_shape}, got {field.shape}")

            if np.isinf(field).any():
                raise ValueError(f"Field `{name}` contains infinities")

            if np.isnan(field).any():
                with_nans.append(name)

        if with_nans:
            msg = f"NaNs found in the following variables: {sorted(with_nans)}"
            if self.allow_nans is None:
                LOG.warning(msg)
                self.allow_nans = True

            if not self.allow_nans:
                raise ValueError(msg)

        return input_state

    def _print_tensor(self, title, tensor_numpy, tensor_by_name, kinds):

        assert len(tensor_numpy.shape) == 3, tensor_numpy.shape
        assert tensor_numpy.shape[0] in (1, self.checkpoint.multi_step_input), tensor_numpy.shape
        assert tensor_numpy.shape[1] == len(tensor_by_name), tensor_numpy.shape

        t = []
        for k, v in enumerate(tensor_by_name):
            data = tensor_numpy[-1, k]

            nans = "-"

            if np.isnan(data).any():
                nan_count = np.isnan(data).sum()

                ratio = nan_count / data.size
                nans = f"{ratio:.0%}"

            if np.isinf(data).any():
                nans = "∞"

            t.append((k, v, np.nanmin(data), np.nanmax(data), nans, kinds.get(v, Kind())))

        LOG.info("")
        LOG.info(
            "%s:\n\n%s\n", title, table(t, header=["Index", "Variable", "Min", "Max", "NaNs", "Kind"], align="><<<|<")
        )
        LOG.info("")

    def _print_input_tensor(self, title, input_tensor_torch):

        input_tensor_numpy = input_tensor_torch.cpu().numpy()  # (batch, multi_step_input, values, variables)

        assert len(input_tensor_numpy.shape) == 4, input_tensor_numpy.shape
        assert input_tensor_numpy.shape[0] == 1, input_tensor_numpy.shape

        input_tensor_numpy = np.squeeze(input_tensor_numpy, axis=0)  # Drop the batch dimension
        input_tensor_numpy = np.swapaxes(input_tensor_numpy, -2, -1)  # (multi_step_input, variables, values)

        self._print_tensor(title, input_tensor_numpy, self._input_tensor_by_name, self._input_kinds)

    def _print_output_tensor(self, title, output_tensor_numpy):

        LOG.info(
            "%s",
            f"Output tensor shape={output_tensor_numpy.shape}, NaNs={np.isnan(output_tensor_numpy).sum()/ output_tensor_numpy.size: .0%}",
        )

        if not self._output_tensor_by_name:
            for i in range(output_tensor_numpy.shape[1]):
                self._output_tensor_by_name.append(self.checkpoint.output_tensor_index_to_variable[i])
                if i in self.checkpoint.prognostic_output_mask:
                    self._output_kinds[self.checkpoint.output_tensor_index_to_variable[i]] = Kind(prognostic=True)
                else:
                    self._output_kinds[self.checkpoint.output_tensor_index_to_variable[i]] = Kind(diagnostic=True)

        # output_tensor_numpy = output_tensor_numpy.cpu().numpy()

        if len(output_tensor_numpy.shape) == 2:
            output_tensor_numpy = output_tensor_numpy[np.newaxis, ...]  # Add multi_step_input

        output_tensor_numpy = np.swapaxes(output_tensor_numpy, -2, -1)  # (multi_step_input, variables, values)

        self._print_tensor(title, output_tensor_numpy, self._output_tensor_by_name, self._output_kinds)



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

from ..context import Context
from ..output import ForwardOutput
from ..output import Output
from . import create_output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("truth")
class TruthOutput(ForwardOutput):
    """Write the truth state for each output state."""

    def __init__(self, context: Context, output: dict, **kwargs):
        super().__init__(context, **kwargs)
        self._input = self.context.create_input()
        self.output: Output = create_output(context, output)

    def write_initial_state(self, state):
        self.output.write_initial_state(state)

    def write_state(self, state):
        truth_state = self._input.create_input_state(date=state["date"])
        reduced_state = self.reduce(truth_state)
        self.output.write_state(reduced_state)

    def write_step(self, state):
        raise NotImplementedError("TruthOutput does not support write_step")

    def open(self, state):
        self.output.open(state)

    def close(self):
        self.output.close()

    def __repr__(self):
        return f"TruthOutput({self.output})"

    def print_summary(self, depth=0):
        super().print_summary(depth)
        self.output.print_summary(depth + 1)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from earthkit.data.utils.dates import to_datetime

from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


def rounded_area(area):
    try:
        surface = (area[0] - area[2]) * (area[3] - area[1]) / 180 / 360
        if surface > 0.98:
            return [90, 0.0, -90, 360]
    except TypeError:
        pass
    return area


def grid_is_valid(grid):
    if grid is None:
        return False

    if isinstance(grid, str):
        return True

    try:
        [float(x) for x in grid]
        return True
    except TypeError:
        return False


def area_is_valid(area):

    if area is None:
        return False

    if len(area) != 4:
        return False

    try:
        [float(x) for x in area]
        return True
    except TypeError:
        return False


def postproc(grid, area):
    pproc = dict()
    if grid_is_valid(grid):
        pproc["grid"] = grid

    if area_is_valid(area):
        pproc["area"] = rounded_area(area)

    return pproc


def retrieve(requests, grid, area, **kwargs):
    import earthkit.data as ekd

    def _(r):
        mars = r.copy()
        for k, v in r.items():
            if isinstance(v, (list, tuple)):
                mars[k] = "/".join(str(x) for x in v)
            else:
                mars[k] = str(v)

        return ",".join(f"{k}={v}" for k, v in mars.items())

    pproc = postproc(grid, area)

    result = ekd.from_source("empty")
    for r in requests:
        if r.get("class") in ("rd", "ea"):
            r["class"] = "od"

        if r.get("type") == "fc" and r.get("stream") == "oper" and r["time"] in ("0600", "1800"):
            r["stream"] = "scda"

        r.update(pproc)
        r.update(kwargs)

        LOG.debug("%s", _(r))

        result += ekd.from_source("mars", r)

    return result


@input_registry.register("mars")
class MarsInput(GribInput):
    """Get input fields from MARS"""

    def __init__(self, context, *, namer=None, **kwargs):
        super().__init__(context, namer=namer)
        self.kwargs = kwargs
        self.variables = self.checkpoint.variables_from_input(include_forcings=False)
        self.kwargs = kwargs

    def create_input_state(self, *, date):
        if date is None:
            date = to_datetime(-1)
            LOG.warning("MarsInput: `date` parameter not provided, using yesterday's date: %s", date)

        date = to_datetime(date)

        return self._create_input_state(
            self.retrieve(
                self.variables,
                [date + h for h in self.checkpoint.lagged],
            ),
            variables=self.variables,
            date=date,
        )

    def retrieve(self, variables, dates):

        requests = self.checkpoint.mars_requests(
            variables=variables,
            dates=dates,
            use_grib_paramid=self.context.use_grib_paramid,
        )

        if not requests:
            raise ValueError("No requests for %s (%s)" % (variables, dates))

        kwargs = self.kwargs.copy()
        kwargs.setdefault("expver", "0001")

        return retrieve(requests, self.checkpoint.grid, self.checkpoint.area, **kwargs)

    def template(self, variable, date, **kwargs):
        return self.retrieve([variable], [date])[0]

    def load_forcings(self, variables, dates):
        return self._load_forcings(self.retrieve(variables, dates), variables, dates)
